#include <cstdio>
#include <string>
#include <vector>
#include <map>
#include <sys/stat.h>
#include <dirent.h>
#include <unistd.h>
#include <fcntl.h>
#include <time.h>
#include <algorithm>

using namespace std;

static map<string, struct stat> stats;
static map<ino_t, string> inomap;
static uid_t uid;
static gid_t gid;
static int filtermin;
static string filterfile;
static string desctemplatestart = "/source/"; // "index/";
static string desctemplateend   = ".html";    // "";

static const string getprog(const string &s)
{
    unsigned startpos=0; 
    if(s.size() > 6 && s[0]=='p' && s[1]=='a' && s[2]=='t' && s[3]=='c' && s[4]=='h' && s[5]=='-')
        startpos = 6;
    unsigned len=0, b=s.size();
    for(unsigned a=startpos; a<b && !(/*s[a]=='.' ||*/ (s[a]=='-' && s[a+1]>='0' && s[a+1]<='9')); ++a)++len;
    //printf("DEBUG:getprog(%s) -> %s\n", s.c_str(), s.substr(startpos,len).c_str());
    return s.substr(startpos, len);
}

static unsigned getvers(const string &s)
{
    bool ispatch = false;
    if(s.size() > 6 && s[0]=='p' && s[1]=='a' && s[2]=='t' && s[3]=='c' && s[4]=='h' && s[5]=='-')
        ispatch = true;
    unsigned result=0, b=s.size();
    vector<unsigned> tokens;
    {unsigned k=1;
    for(unsigned a=0; a<b; ++a)
    {
        if(s[a] >= '0' && s[a] <= '9')
        {
            if(a>0 && (s[a-1]>='a' && s[a-1]<='z'))goto Alfa;
            //printf("DEBUG: result=%u*%u + %c\n", result,k, s[a]);
            result = result*k + (s[a]-'0');
            k=10;
        }
        else
        {
        Alfa:
            if(k==10)
            {
                tokens.push_back(result); result=0;
                if(s[a]=='-' && ispatch)break; k=1;
            }
        }
    }}
    tokens.push_back(result); result=0;
    for(unsigned a=0, k=20*200*200*400; a<tokens.size(); ++a)
    {
        result += tokens[a] * k;
        switch(a)
        {
            case 0: k /= 20; break;
            case 1: k /= 200; break;
            case 2: k /= 200; break;
            case 3: k /= 400; break;
        }
    }
    result*=2;
    if(ispatch)++result;
    //printf("DEBUG:getvers(%s) -> %u\n", s.c_str(), result);
    return result;
}

class com
{
public:
    const bool operator() (const string &a1, const string &b1) const
    {
        const string a = getprog(a1);
        const string b = getprog(b1);
        if(a < b)return true;
        if(a > b)return false;
        return (getvers(a1) > getvers(b1));
    }
};

static const string urlencode(const string &s)
{
    string result;
    for(unsigned a=0; a<s.size(); ++a)
    {
        if((unsigned char)s[a] >= 0x7E
        || s[a] == '%'
        || s[a] == '\\'
        || s[a] == ' ')
        {
            char Buf[4];
            sprintf(Buf, "%%%02X", (unsigned char)s[a]);
            result += Buf;
        }
        else result += s[a];
    }
    return result;
}

static const string htmlspecialchars(const string &s)
{
    string result;
    for(unsigned a=0; a<s.size(); ++a)
    {
        if(s[a]=='&')result += "&amp;";
        else if(s[a]=='"')result += "&quot;";
        else if(s[a]=='<')result += "&lt;";
        else result += s[a];
    }
    return result;
}

static string basedir;

static string Relative(const string &base, const string &dest)
{
    if(dest[0] == '/')
        return dest;
    
    string res = base;

    /* Eat the rightmost item to get the directory */
    unsigned a = res.rfind('/');
    if(a != res.npos)
        res = res.substr(0, a + 1); /* include / */
    else
        res = "";

    res += dest;
    
    unsigned b = 0;
    for(;;)
    {
        a = res.find("/.", b);
        if(a == res.npos)break;
        if(a+2 < res.size() && res[a+2] == '.')
        {
            if(a+3 < res.size() && res[a+3] != '/')
            {
                /* No hit */
                b = a+1;
                continue;
            }
            /* Todo: Erase one level */
            b = a+1;
            continue;
            continue;
        }
        if(a+2 < res.size() && res[a+2] != '/')
        {
            b = a+1;
            continue;
        }

        /* Omit /.(/) */
        res = res.substr(0, a) + res.substr(a+2);
    }
    return res;
}

static void display(const string &s, bool isdesc = false)
{
    getprog(s);
    getvers(s);
    
    struct stat st = stats[s];
    struct tm *t = localtime(&st.st_mtime);
    
    unsigned mode = st.st_mode;
    ino_t inode = st.st_ino;
    
    bool link_ok = true;
    if(S_ISLNK(mode))
    {
        /* Find link target stats */
        if(stat(s.c_str(), &st) < 0)
        {
            /* Link is invalid */
            link_ok = false;
        }
    }
    
    printf(
        "\n"
        "%04d-%02d%02d-%02d%02d ",
        t->tm_year+1900,
        t->tm_mon+1, t->tm_mday,
        t->tm_hour, t->tm_min);
    if(link_ok)
    {
        printf(
            "<small>%c%c%c</small>"
            "%10lu",
            (((st.st_mode >> (uid==st.st_uid?6:(gid==st.st_gid?3:0))) & 4) ? 'r' : '-'),
            (((st.st_mode >> (uid==st.st_uid?6:(gid==st.st_gid?3:0))) & 2) ? 'w' : '-'),
            (((st.st_mode >> (uid==st.st_uid?6:(gid==st.st_gid?3:0))) & 1) ? 'x' : '-'),
            (unsigned long)st.st_size
        );
    }
    else
    {
        printf("<small>---</small>%10s", "-");
    }
    printf(" <a href=\"");
    string url = basedir + (isdesc ? desctemplatestart + s + desctemplateend : s);

    if(!isdesc
    && (st.st_size/1024 >= filtermin)
    && filterfile.size()
    && !S_ISDIR(st.st_mode)
    )
    {
        url = filterfile + url;
    }
    
    printf("%s", htmlspecialchars(urlencode(url)).c_str());
    if(S_ISDIR(st.st_mode))putchar('/');
    printf("\">%s", htmlspecialchars(s).c_str());
    if(isdesc)printf(".html");
    printf("</a>");
    if(S_ISDIR(st.st_mode))putchar('/');
    
    if(S_ISLNK(mode))
    {
        string Target = s;
        /* Loop to find the link path */
        for(;;)
        {
        	const string Base = Target;
        	
            char Buf[1024];
            memset(Buf,0,sizeof Buf);
            readlink(Target.c_str(), Buf, sizeof Buf);

            Target = Relative(Base, Buf);
        
            printf(" <font color=blue>-&gt;</font> <font color=\"#806090\">%s</font>",
                   htmlspecialchars(Target).c_str());
            
            struct stat st1;
            if(lstat(Target.c_str(), &st1) < 0)
            {
                /* Found a broken link */
                printf(" <font color=\"#906030\">?</font>");
                break;
            }
            if(!S_ISLNK(st1.st_mode))
            {
                /* Found a nonlink */
                break;
            }
        }
    }
    if(isdesc)
    {
        printf("<font color=blue>@</font>");
        FILE *fp = fopen((".desc-" + s + ".php").c_str(), "rt");
        if(fp)
        {
            for(;;)
            {
                char Buf[256];
                if(!fgets(Buf, sizeof Buf, fp))break;
                if(strncmp(Buf, "//TITLE=", 8))continue;
                strtok(Buf, "\r\n");
                printf(" (<font color=\"#500010\">%s</font>)", htmlspecialchars(Buf+8).c_str());
                break;
            }
            fclose(fp);
        }
    }
    
    map<ino_t, string>::const_iterator i = inomap.find(inode);
    if(i == inomap.end())
        inomap[inode] = s;
    else if(i->second != s)
        printf(" <font color=blue>=</font> <font color=\"#806090\">%s</font>",
               htmlspecialchars(i->second).c_str());
}

static vector<string> inclusions;

#include "argh.hh"

int main(int argc, const char *const *argv)
{
    bool descspecial = false;
    string dir = ".";

    /* Parameter handling begins */
    ParamHandler Argh;
    Argh.AddLong("help",     'h').SetBool().SetDesc("This help.");
    Argh.AddLong("version",  'V').SetBool().SetDesc("Displays version information.");
    Argh.AddLong("directory",'d').SetString().SetDesc("Directory (local) to list.", "<dir>");
    Argh.AddLong("basedir",  'b').SetString().SetDesc("Directory (virtual) to refer to (end with /).", "<dir>");
    Argh.AddLong("filter",   'f').SetString().SetDesc("Filter pathname for file downloads.", "<phpfile>");
    Argh.AddLong("filtersize",'s').SetInt(0,0x7FFFFFFF).SetDesc("Minimum filesize for filtering in kB.", "<int>");
    Argh.AddLong("descspecial",'c').SetBool().SetDesc("Handle .desc*.php files specially");
    Argh.StartParse(argc, argv);
    for(;;)
    {
        int c = Argh.GetParam();
        if(c == -1)break;
        switch(c)
        {
            case 'V': printf("%s\n", VERSION); return 0;
            case 'h':
                printf(
                    "dirlister v"VERSION" - Copyright (C) 1992,2001 Bisqwit (http://iki.fi/bisqwit/)\n"
                    "\nUsage: dirlister [<option> [<...>]] [<prog> [<...>]]\n"
                    "\nOutputs the file listing in a directory as html.\n"
                    "\nOptions:\n");
                Argh.ListOptions();
                printf("\nNo warranty whatsoever.\n");
                return 0;
            case 'd': dir = Argh.GetString(); break;
            case 'c': descspecial = Argh.GetBool(); break;
            case 'b': basedir = Argh.GetString(); break;
            case 's': filtermin = Argh.GetInt(); break;
            case 'f': filterfile= Argh.GetString(); break;
            default:
                inclusions.push_back(Argh.GetString());
        }
    }
    if(!Argh.ok())return -1;
    /* Parameter handling ends */
    
    vector<string> files, dirs, descs;
    
    chdir(dir.c_str());
    
    DIR *d = opendir(".");
    struct dirent *ent;
    while((ent = readdir(d)))
    {
        string d(ent->d_name);
        
        if(inclusions.size())
        {
            for(unsigned a=0; a<inclusions.size(); ++a)
                if(getprog(d) == inclusions[a])goto Found;
            continue;
        Found: ;
        }
        
        if(descspecial && d.substr(0, 6) == ".desc-")
        {
            const string entnam = d.substr(6, d.size() - 6 - 4);
            descs.push_back(entnam);   /* 6+4 = 10 = ".desc-" + ".php" */
            struct stat diu;
            lstat(ent->d_name, &diu);
            stats[entnam] = diu;
            continue;
        }
        
        if(d[0] == '.' && d != "..")continue;

        struct stat diu;
        lstat(ent->d_name, &diu);
        stats[d] = diu;
        if(S_ISDIR(diu.st_mode))
            dirs.push_back(d);
        else
            files.push_back(d);
    }
    closedir(d);
    
    sort(descs.begin(), descs.end(), com());
    sort(files.begin(), files.end(), com());
    sort(dirs.begin(),  dirs.end(),  com());
    
    printf(
        "<PRE class=filelist>"
        "Date (Y-md-Hi) <small>acc</small>      Size Name\n"
        "-----------------------------------------------------");

    uid = getuid();
    gid = getgid();
    
    vector<string>::const_iterator i;
    for(i=dirs.begin();  i!=dirs.end(); ++i)display(*i);
    for(i=descs.begin(); i!=descs.end(); ++i)display(*i, true);
    for(i=files.begin(); i!=files.end(); ++i)
    {
        if(descspecial)
        {
            const string prog = getprog(*i);
            for(unsigned a=0; a<descs.size(); ++a)
                if(descs[a] == prog)goto Bad;
        }
        display(*i);
    Bad: ;
    }
    
    printf("</PRE>\n");
    
    return 0;
}
